/* Copyright (C) 2005 Juergen "George" Sawinski
 * All rights reserved.
 * Licensed under the LGPL v2.1 or later,
 * see accompanying COPYING.LIB file.
 */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <stdarg.h>
#include <assert.h>
#include <unistd.h>
#include <stdlib.h>
#include <limits.h>

#include <grp.h>
#include <pwd.h>

#include "tar.h"

/*-----------------------------------------------------------------------------*\
| helpers                                                                       |
\*-----------------------------------------------------------------------------*/
static tar_ud *createhandle(lua_State *L)
{
  tar_ud *ud;

  ud = (tar_ud *) lua_newuserdata(L, sizeof(tar_ud));
  
  /* initialize */
  ud->L = L;
  ud->ref  = LUA_REFNIL;
  ud->file = NULL;

  /* set metatable */
  luaL_getmetatable(L, LibType);
  lua_setmetatable(L, -2);
  
  /* create and store reference to callback table */
  luaL_getmetatable(L, LibType);
  ud->ref = luaL_ref(L, LUA_REGISTRYINDEX);

  return ud;
}

static tar_ud *gethandle(lua_State *L)
{
  tar_ud *ud;
    
  /* get and check first argument (should be tar) */
  ud = (tar_ud *)luaL_checkudata(L, 1, LibType);
  luaL_argcheck(L, ud, 1, "Tar file handled expected.");

  /* update state */
  ud->L = L;

  return ud;
}

static void setstring(lua_State *L, const char *key, const char *value)
{
  lua_pushstring(L, key);
  lua_pushstring(L, value);
  lua_settable(L, -3);
}

static void setint(lua_State *L, const char *key, long value)
{
  lua_pushstring(L, key);
  lua_pushinteger(L, value);
  lua_settable(L, -3);
}

/*-----------------------------------------------------------------------------*\
| open,close                                                                    |
\*-----------------------------------------------------------------------------*/
/** handle, error = tar.open("pathname")
 ** handle, error = tar.open{pathname = "pathname" [, type = filetype] }
 *    
 *  Open a tar file at "pathname" for reading. Optionally, if the file type cannot
 *  be autodetected by its file magic, a filetype argument can be passed, see "tar.TYPE_*"
 *
 *  Returns: tar object handle on success, otherwise nil and the error message.
 */
static int luatar_open(lua_State *L)
{
  const char *fname = NULL;
  tar_ud *ud;
  int type = -1;
  int flags = TAR_GNU;

  /* function arguments */
  if (lua_type(L, 1) == LUA_TTABLE) {
    lua_pushnil(L);  /* first key */
    while (lua_next(L, 1) != 0) {
      const char *key = luaL_checkstring(L,-2);

      dbg("=========== open(key = %s)\n", key);
      
      if (!strcmp(key, "pathname")) {
	if (lua_isstring(L, -1)) 
	  fname = lua_tostring(L, -1);
      }
      else if (!strcmp(key, "type")) {
	if (lua_isnumber(L, -1)) 
	  type = lua_tointeger(L, -1);
	else if (lua_isstring(L, -1)) {
	  type = luatar_filetype(lua_tostring(L, -1));
	}
      }
      else if (!strcmp(key, "flags")) {
	if (lua_isnumber(L, -1)) 
	  flags = lua_tointeger(L, -1);
      }

      lua_pop(L, 1);
    }
  }
  else if (lua_type(L, 1) == LUA_TSTRING) {
    fname = luaL_optlstring(L, 1, NULL, NULL);
  }
  else {
    /* FIXME report error */
    return 0;
  }

  /* no type? then guess */
  if (type < 0) type = luatar_filemagic(fname);
  if (type < 0) type = luatar_filetype(fname);

  /* FIXME error if no type */

  /* create handle and open file */
  ud = createhandle(L);

  if (aux_luatar_open(ud, fname, type, flags)) {
    lua_pop(ud->L, 1); /* FIXME cleanup userdata */

    /* return error */
    /* TODO sort out to error function */
    lua_pushnil(ud->L);
    lua_pushfstring(ud->L, "%s (%s)", strerror(errno), fname);

    return 2;
  }

  return 1;
}

/** handle,error = tar.new{pathname = "pathname", [options] }
 *    
 *  Open a tar file at "pathname" for writing.
 *
 *  Options:
 *    type = "type" or tar.TYPE.*
 *    compression = level
 *    mode = unix file mode
 *    flags = tar flags (see tar.TAR.*)
 *
 *  Returns: tar object handle on success, otherwise nil and the error message.
 */
static int luatar_new(lua_State *L)
{
  tar_ud *ud;

  const char *fname = NULL;
  int type = LUATAR_TYPE_PLAIN;
  int compression = 9;
  int flags = TAR_GNU;
  int mode = 0644;

  dbg("luatar_new(...)\n");

  /* argument must be a table */
  luaL_checktype(L, 1, LUA_TTABLE);
  
  /* function arguments */
  lua_pushnil(L);  /* first key */
  while (lua_next(L, 1) != 0) {
    const char *key = luaL_checkstring(L,-2);
    
    if (!strcmp(key, "pathname")) {
      if (lua_isstring(L, -1)) {
	fname = lua_tostring(L, -1);
      }
    }
    else if (!strcmp(key, "type")) {
      if (lua_isnumber(L, -1)) 
	type = lua_tointeger(L, -1);
      else if (lua_isstring(L, -1)) {
	type = luatar_filetype(lua_tostring(L, -1));
      }
    }
    else if (!strcmp(key, "compression")) {
      if (lua_isnumber(L, -1)) 
	compression = lua_tointeger(L, -1);
    }
    else if (!strcmp(key, "flags")) {
      if (lua_isnumber(L, -1)) 
	flags = lua_tointeger(L, -1);
    }
    else if (!strcmp(key, "mode")) {
      if (lua_isnumber(L, -1)) 
	mode = lua_tointeger(L, -1);
      else if (lua_isstring(L, -1))
	mode = oct_to_int((char *)/* FIXME why that? */ lua_tostring(L,-1));
    }
    else {
      /* ignore */
      /* FIXME print warning */
    }
    
    lua_pop(L, 1);
  }

  /* create handle and open file */
  ud = createhandle(L);

  if (aux_luatar_new(ud, fname, type, compression, flags, mode)) {
    lua_pop(ud->L, 1); /* FIXME cleanup userdata */

    /* return error */
    /* TODO sort out to error function */
    lua_pushnil(ud->L);
    lua_pushfstring(ud->L, "%s (%s)", strerror(errno), fname);

    return 2;
  }

  return 1;
}

/** tar:close()
 ** tar.close(handle)
 *
 *  Close the tar file.
 */
static int luatar_close(lua_State *L)
{
  tar_ud *ud = gethandle(L);

  /* object still referenced ? */
  if (ud && (ud->ref != LUA_REFNIL)) {
    /* free (unref) callback table */
    luaL_unref(L, LUA_REGISTRYINDEX, ud->ref);
    ud->ref = LUA_REFNIL;
  
    /* close tar file */
    aux_luatar_close(ud);
  }
  
  return 0;
}

/*-----------------------------------------------------------------------------*\
| object methods                                                                |
\*-----------------------------------------------------------------------------*/
static int __luatar_header_iterator(lua_State *L)
{
  tar_ud *ud = (tar_ud *) lua_topointer (L, lua_upvalueindex(1));

  dbg("__luatar_header_iterator()\n");
  dbg("    status: %d\n", ud->status);

  /* return due to error */
  if (ud->status == LUATAR_STATUS_ERROR) return 0;

  /* file was not handled */
  if (ud->status == LUATAR_STATUS_HEADER) tar_skip_regfile(ud->file);

  /* read header */
  if (th_read(ud->file)) {
    ud->status = LUATAR_STATUS_ERROR;
    return 0;
  }
  ud->status = LUATAR_STATUS_HEADER;

  /* push file name */
  lua_pushstring(L, th_get_pathname(ud->file));

  return 1;
}

/** iterator = tar:files()
 *
 *  Create iterator over file contents.
 *
 *  Example:
 *    t = tar.open("some-file");
 *    for f in t:files() do
 *      print(f)
 *      local info = t:info()
 *      for k,v in next,info do
 *        print("", k, v)
 *      end
 *    end
 *
 *  Returns: iterator (returning file name) over tar file contents
 */
static int luatar_files(lua_State *L)
{
  tar_ud *ud = gethandle(L);

  if (!ud->read_only) return 0;

  /* push iterator */
  lua_pushlightuserdata (L, ud);
  lua_pushcclosure (L, __luatar_header_iterator, 1);
  return 1;
}

/** table = tar:info()
 *
 *  Get the header info.
 *
 *  Returns: the file header info as table.
 */
static int luatar_info(lua_State *L)
{
  tar_ud *ud = gethandle(L);

  /* return due to error */
  if (ud->status == LUATAR_STATUS_ERROR) return 0;

  /* not initialized! */
  if (ud->status == LUATAR_STATUS_INIT) {
    if (!ud->read_only) return 0;
    
    /* read header */
    if (th_read(ud->file)) {
      ud->status = LUATAR_STATUS_ERROR;
      return 0;
    }

    ud->status = LUATAR_STATUS_HEADER;
  }

  /* put header into a table */
  lua_newtable(L);

  /* extract header */
  setstring(L, "pathname", th_get_pathname(ud->file));
  setstring(L, "linkname", th_get_linkname(ud->file));
  setint(L, "mode",     th_get_mode(ud->file)); /* FIXME use int_to_oct */
  setint(L, "uid",      th_get_uid(ud->file));
  setint(L, "gid",      th_get_gid(ud->file));
  setint(L, "crc",      th_get_crc(ud->file));
  setint(L, "size",     th_get_size(ud->file));
  setint(L, "mtime",    th_get_mtime(ud->file));
  setint(L, "devmajor", th_get_devmajor(ud->file));
  setint(L, "devminor", th_get_devminor(ud->file));

  /* file type */
  {
    const char *filetype = "";

    if (TH_ISREG(ud->file)) filetype = "reg";
    else if (TH_ISLNK(ud->file)) filetype = "link";
    else if (TH_ISSYM(ud->file)) filetype = "sym";
    else if (TH_ISCHR(ud->file)) filetype = "char";
    else if (TH_ISBLK(ud->file)) filetype = "block";
    else if (TH_ISDIR(ud->file)) filetype = "dir";
    else if (TH_ISFIFO(ud->file)) filetype = "fifo";
    
    setstring(L, "type", filetype);
  }

  return 1;
}

/* macros for reading/writing tarchive blocks (see libtar.h) */
#define luatar_block_read(t, buf, len)				\
  (*((t)->type->readfunc))((t)->fd, (char *)(buf), len)
#define luatar_block_write(t, buf, len)				\
  (*((t)->type->writefunc))((t)->fd, (char *)(buf), len)


/** content,errro = tar:read()
 *
 *  Read the contents of a file.
 *
 *  Returns: contents of a file
 */
static int luatar_read(lua_State *L)
{
  tar_ud *ud = gethandle(L);
  ptrdiff_t size;
  char *buf;
  char *ptr;

  dbg("luatar_read()\n");

  /* the header must be read */
  if (ud->status != LUATAR_STATUS_HEADER) {
    /* FIXME report error */
    return 0;
  }

  /* create buffer */
  /* FIXME use a static buffer ala lzlib.c */
  size = th_get_size(ud->file);
  ptr = buf = malloc(size+T_BLOCKSIZE-1);

  /* read file contents */
  while(size > 0) {
    int read_bytes = tar_block_read(ud->file, ptr);
    if (read_bytes <= 0) break;
    
    ptr  += read_bytes;
    size -= read_bytes;
  }
  ud->status = LUATAR_STATUS_FILE;

  /* return result */
  lua_pushlstring(L, buf, th_get_size(ud->file));

  /* FIXME used a static buffer, see lzlib.c */
  free(buf);

  return 1;
}

/** success,error = tar:extract()
 *  success,error = tar:extract(realname)
 *  success,error = tar:extract{
 *	path     = "path-prefix",
 *      realname = "real-name",
 *      gid      = number,
 *      uid      = number,
 *      mtime    = mtime
 *  }
 *
 *  Extract a file from tar file.
 */
static int luatar_extract(lua_State *L)
{
  tar_ud *ud = gethandle(L);
  char *realname = alloca(PATH_MAX + 1);
  gid_t gid;
  uid_t uid;
  time_t mtime;

  dbg("luatar_extract(...)\n");

  /* check status */
  if (ud->status != LUATAR_STATUS_HEADER) {
    /* FIXME report warning or error ? */
    return 0;
  }

  /* initialize */
  gid = th_get_gid(ud->file);
  uid = th_get_uid(ud->file);
  mtime = th_get_mtime(ud->file);
  snprintf(realname, PATH_MAX, th_get_pathname(ud->file));

  /* parse arguments realname */
  if (lua_type(L, 2) == LUA_TTABLE) {
    int have_realname = 0;

    lua_pushnil(L);  /* first key */
    while (lua_next(L, 2) != 0) {
      const char *key = luaL_checkstring(L,-2);

      if (!strcmp(key, "realname")) {
	snprintf(realname, PATH_MAX, "%s", lua_tostring(L, -1));
	have_realname = 1;
      }
      else if (!strcmp(key, "path")) {
	if (!have_realname) {
	  size_t n = snprintf(realname, PATH_MAX, "%s", lua_tostring(L, -1));
	  /* FIXME auto-append "/" if missing ? */
	  snprintf(realname+n, PATH_MAX-n, "%s", th_get_pathname(ud->file));
	}
      }
      else if (!strcmp(key, "gid")) {
	gid = lua_tointeger(L, -1);
      }
      else if (!strcmp(key, "uid")) {
	uid = lua_tointeger(L, -1);
      }
      else if (!strcmp(key, "mtime")) {
	mtime = lua_tointeger(L, -1);
      }

      lua_pop(L, 1);
    }
  }
  else if (lua_type(L, -1) == LUA_TSTRING) {
    snprintf(realname, PATH_MAX, "%s", lua_tostring(L, 1));
  }
  else {
    /* FIXME ignore only if no argument */
  }

  /* adjust uid,gid and mtime */
  th_set_group(ud->file, gid);
  th_set_user(ud->file, uid);
  th_set_mtime(ud->file, mtime);

  /* extract */
  if (tar_extract_file(ud->file, realname)) {
    /* FIXME report error */
    ud->status = LUATAR_STATUS_ERROR;
    return 0;
  }

  ud->status = LUATAR_STATUS_FILE;
  lua_pushboolean(L, 1);
  return 1;
}


/*-----------------------------------------------------------------------------*\
| Library init                                                                  |
\*-----------------------------------------------------------------------------*/
static const struct luaL_reg methods[] = {
  {"files",   luatar_files },
  {"info",    luatar_info },

  /* reading */
  {"read",    luatar_read },
  {"extract", luatar_extract },

  /* manual and auto-close */
  {"close",   luatar_close},
  {"__gc",    luatar_close},
  /*{"__tostring", lua_tar_tostring}*/
  {NULL, NULL}
};

static const struct luaL_reg functions[] = {
  {"new",      luatar_new},
  {"open",     luatar_open},
  {NULL, NULL}
};

LUALIB_API int LibInitFunc (lua_State *L) {
  /* create metatable */
  luaL_newmetatable(L, LibType);
  
  /* metatable.__index = metatable */
  lua_pushliteral(L, "__index");
  lua_pushvalue(L, -2);
  lua_settable(L, -3);

  /* register functions and methods methods */
  luaL_openlib (L, NULL, methods, 0);
  luaL_openlib (L, LibName, functions, 0);

  /* register type flags */
  lua_newtable(L);

  lua_pushliteral(L, "TYPE");
  lua_pushvalue(L, -2);
  lua_settable(L, -4);

#define TYPEFLAG(flag)				\
  lua_pushliteral(L, #flag);			\
  lua_pushinteger(L, LUATAR_TYPE_##flag);	\
  lua_settable(L, -3);

  TYPEFLAG(PLAIN);
  TYPEFLAG(ZIP);
  TYPEFLAG(GZIP);
  TYPEFLAG(BZIP2);

  lua_pop(L, 1);

  /* register tar flags */
  lua_newtable(L);

  lua_pushliteral(L, "TAR");
  lua_pushvalue(L, -2);
  lua_settable(L, -4);

#define TARFLAG(flag)				\
  lua_pushliteral(L, #flag);			\
  lua_pushinteger(L, TAR_##flag);		\
  lua_settable(L, -3);

  TARFLAG(GNU);
  TARFLAG(VERBOSE);
  TARFLAG(NOOVERWRITE);
  TARFLAG(IGNORE_EOT);
  TARFLAG(IGNORE_MAGIC);
  TARFLAG(CHECK_VERSION);
  TARFLAG(IGNORE_CRC);

  lua_pop(L, 1);
  
  return 1;
}

